use std::pin::Pin;

use super::{util::TaggedAtomicPtr, *};

fn pin_ptr<T>(pin: &Pin<&mut T>) -> NonNull<T> {
    let a = &raw const **pin;
    unsafe { NonNull::new_unchecked(a.cast_mut()) }
}

#[test]
fn job_list_pop_back() {
    let mut list = JobList::new();
    let mut a = pin!(Job::empty());
    let mut b = pin!(Job::empty());
    let mut c = pin!(Job::empty());

    unsafe {
        list.push_front(a.as_ref());
        list.push_front(b.as_ref());
        list.push_back(c.as_ref());
    }

    assert_eq!(list.pop_back(), Some(pin_ptr(&c)));
    unsafe {
        list.push_front(c.as_ref());
    }
    assert_eq!(list.pop_back(), Some(pin_ptr(&a)));
    assert_eq!(list.pop_back(), Some(pin_ptr(&b)));
    assert_eq!(list.pop_back(), Some(pin_ptr(&c)));
    assert_eq!(list.pop_back(), None);
    assert_eq!(list.pop_front(), None);
}

#[test]
fn job_list_pop_front() {
    let mut list = JobList::new();
    let mut a = pin!(Job::<()>::empty());
    let mut b = pin!(Job::<()>::empty());
    let mut c = pin!(Job::<()>::empty());

    unsafe {
        list.push_front(a.as_ref());
        list.push_front(b.as_ref());
        list.push_back(c.as_ref());
    }

    assert_eq!(list.pop_front(), Some(pin_ptr(&b)));
    unsafe {
        list.push_back(b.as_ref());
    }
    assert_eq!(list.pop_front(), Some(pin_ptr(&a)));
    assert_eq!(list.pop_front(), Some(pin_ptr(&c)));
    assert_eq!(list.pop_front(), Some(pin_ptr(&b)));
    assert_eq!(list.pop_front(), None);
    assert_eq!(list.pop_back(), None);
}

#[test]
fn unlink_job_middle() {
    let mut list = JobList::new();
    let mut a = pin!(Job::<()>::empty());
    let mut b = pin!(Job::<()>::empty());
    let mut c = pin!(Job::<()>::empty());

    unsafe {
        list.push_front(a.as_ref());
        list.push_front(b.as_ref());
        list.push_front(c.as_ref());
    }

    unsafe {
        b.unlink();
    }

    assert_eq!(list.pop_front(), Some(pin_ptr(&c)));
    assert_eq!(list.pop_front(), Some(pin_ptr(&a)));
    assert_eq!(list.pop_front(), None);
    assert_eq!(list.pop_back(), None);
}

#[test]
fn unlink_job_front() {
    let mut list = JobList::new();
    let mut a = pin!(Job::<()>::empty());
    let mut b = pin!(Job::<()>::empty());
    let mut c = pin!(Job::<()>::empty());

    unsafe {
        list.push_front(a.as_ref());
        list.push_front(b.as_ref());
        list.push_front(c.as_ref());
    }

    unsafe {
        a.unlink();
    }

    assert_eq!(list.pop_front(), Some(pin_ptr(&c)));
    assert_eq!(list.pop_front(), Some(pin_ptr(&b)));
    assert_eq!(list.pop_front(), None);
    assert_eq!(list.pop_back(), None);
}

#[test]
fn unlink_job_back() {
    let mut list = JobList::new();
    let mut a = pin!(Job::<()>::empty());
    let mut b = pin!(Job::<()>::empty());
    let mut c = pin!(Job::<()>::empty());

    unsafe {
        list.push_front(a.as_ref());
        list.push_front(b.as_ref());
        list.push_front(c.as_ref());
    }

    unsafe {
        c.unlink();
    }

    assert_eq!(list.pop_front(), Some(pin_ptr(&b)));
    assert_eq!(list.pop_front(), Some(pin_ptr(&a)));
    assert_eq!(list.pop_front(), None);
    assert_eq!(list.pop_back(), None);
}

#[test]
fn unlink_job_single() {
    let mut list = JobList::new();
    let a = pin!(Job::<()>::empty());

    unsafe {
        list.push_front(a.as_ref());
    }

    unsafe {
        a.unlink();
    }

    assert_eq!(list.pop_front(), None);
    assert_eq!(list.pop_back(), None);
}

#[test]
fn tagged_ptr_exchange() {
    let boxed = Box::into_raw(Box::new(42usize));
    let ptr = TaggedAtomicPtr::<_, 3>::new(boxed, 1usize);

    assert_eq!(ptr.tag(Ordering::Relaxed), 1);
    assert_eq!(ptr.ptr(Ordering::Relaxed).as_ptr(), boxed);
    ptr.set_tag(3, Ordering::Release, Ordering::Relaxed);
    assert_eq!(ptr.tag(Ordering::Relaxed), 3);
    assert_eq!(ptr.ptr(Ordering::Relaxed).as_ptr(), boxed);
    let old = ptr.compare_exchange_weak_tag(3, 4, Ordering::Release, Ordering::Relaxed);
    assert_eq!(old, Ok(3));
    assert_eq!(ptr.tag(Ordering::Relaxed), 4);
    assert_eq!(ptr.ptr(Ordering::Relaxed).as_ptr(), boxed);
}

#[test]
fn tagged_ptr_exchange_failure() {
    let boxed = Box::into_raw(Box::new(42usize));
    let ptr = TaggedAtomicPtr::<_, 3>::new(boxed, 1usize);

    assert_eq!(ptr.tag(Ordering::Relaxed), 1);
    assert_eq!(ptr.ptr(Ordering::Relaxed).as_ptr(), boxed);

    let old = ptr.compare_exchange_weak_tag(3, 4, Ordering::Release, Ordering::Relaxed);
    assert_eq!(old, Err(1));
    assert_eq!(ptr.tag(Ordering::Relaxed), 1);
    assert_eq!(ptr.ptr(Ordering::Relaxed).as_ptr(), boxed);
}
